/* ----------------------------------------------------
                          uni51_m8.c

     ISP-Flasher fuer MCS51 MCU
     Unterstuetzte Controller:

         AT89S51
         AT89S52
         AT89S8253
         AT89S2051
         AT89S4051

     Controller: ATmega48 - ATmega328; ATmega8


     10.08.2018    R. Seelig

   ---------------------------------------------------- */


#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

// ----------------------------------------------------
// Portpins der UART-USB Bruecke
// ----------------------------------------------------
#define serportddr     DDRD
#define serport        PORTD
#define serpin         PIND
#define serrxd         PD0
#define sertxd         PD1

#define seroutrxd      PD4
#define serintxd       PD7

#define settxd()       ( serport |= (1 << sertxd) )
#define clrtxd()       ( serport &= ~( 1 << sertxd) )

#define setoutrxd()    ( serport |= (1 << seroutrxd) )
#define clroutrxd()    ( serport &= ~( 1 << seroutrxd) )

#define isrxd()        (((serpin) & (1 << serrxd)) >> serrxd)

#define isserintxd()   (((serpin) & (1 << serintxd)) >> serintxd)

// ----------------------------------------------------
// UART Button (Aktivierung / Deaktivierung UART-USB bridge)
// ----------------------------------------------------
#define button_ddr     DDRB
#define button_port    PORTB
#define button         PB0
#define button_in      PINB

#define button_init()  { button_ddr= (1 << button); button_port= (1 << button); }
#define is_button()    (!(((button_in) & (1 << button)) >> button))

// ----------------------------------------------------
// gruene LED zeigt Flashaktivitaet an
// ----------------------------------------------------
#define led_gn           PC2
#define led_gn_dir       DDRC
#define led_gn_port      PORTC

#define led_gn_init()    (led_gn_dir |= 1 << led_gn)
#define led_gn_set()     (led_gn_port |= 1 << led_gn)
#define led_gn_clr()     (led_gn_port &= ~(1 << led_gn))

// ----------------------------------------------------
// gelbe LED als Programmer ready Anzeige
// ----------------------------------------------------
#define led_ge           PC5
#define led_ge_dir       DDRC
#define led_ge_port      PORTC

#define led_ge_init()    (led_gn_dir |= 1 << led_ge)
#define led_ge_set()     (led_gn_port |= 1 << led_ge)
#define led_ge_clr()     (led_gn_port &= ~(1 << led_ge))

// ----------------------------------------------------
// rote LED als Anzeige fuer USB-UART bridge aktiv
// ----------------------------------------------------
#define led_rt           PD5
#define led_rt_dir       DDRD
#define led_rt_port      PORTD

#define led_rt_init()    (led_rt_dir |= 1 << led_rt)
#define led_rt_set()     (led_rt_port |= 1 << led_rt)
#define led_rt_clr()     (led_rt_port &= ~(1 << led_rt))

// ----------------------------------------------------
// globale Variable
// ----------------------------------------------------
char    rinv      = 0;
char    clkinv    = 0;
char    aktfreq   = 3;
char    mcutyp    = 0xff;
int     pagesize;

uint16_t flashbytes;
char     pgwr_wait, bywr_wait, erase_wait;
char     bytebuff[256];

// ----------------------------------------------------
// Controllerspezifisches (Timing und Pagespeichergroesse)
// ----------------------------------------------------
#define page8253     64
#define page2051     32

#define p_delay   4700          // Delay pagewrite in uS
#define p_delay2  4700          // Delay pagewrite in uS
#define b_delay     50          // Delay bytewrite in uS
#define b_delay2  1500          // Delay bytewrite in uS (80S2051 / S4051)
#define e_delay    420          // Delay erase in mS


// ----------------------------------------------------
// Variable fuer den "SPI-Monitor"
// ----------------------------------------------------
#define cmdanz   15
static const char PROGMEM cmds [cmdanz][9] =

//     1          2        3       4        5       6
  { "disable", "enable", "mosi", "miso", "reset", "clk",       \
//     7      8       9       10      11      12       13
    "rinv", "send", "read", "cmd3", "freq", "dump", "clkinv",  \
//    14      15
    "cmd4", "exit" };


/* --------------------------------------------------
            SPI und UART Hardwarefunktionen
   -------------------------------------------------- */

  #if defined (__AVR_ATmega8__) || defined (__AVR_ATtiny2313__)  || defined (__AVR_ATmega8515__)

    #define UBRR0       UBRR
    #define UBRR0L      UBRRL
    #define UBRR0H      UBRRH
    #define UCSR0A      UCSRA
    #define UCSR0B      UCSRB
    #define UCSR0C      UCSRC
    #define U2X0        U2X
    #define RXEN0       RXEN
    #define TXEN0       TXEN
    #define UCSZ00      UCSZ0
    #define UCSZ01      UCSZ1
    #define UDRE0       UDRE
    #define UDR0        UDR
    #define RXC0        RXC

  #endif

  #define  uart_prints(tx)   uart_putromstring(PSTR(tx))         // Benutzung: uart_prints("Hallo Welt\n\r");

  /* --------------------------------------------------
       Zurordnung der Pinbelegung der SPI Schnitt-
       stelle
     -------------------------------------------------- */

  #define SPI_DDR         DDRB
  #define SPI_PORT        PORTB
  #define SPI_INPUT       PINB
  #define SPI_MOSI_PIN    PB3             // Arduino digital 11
  #define SPI_MISO_PIN    PB4             //      "          12
  #define SPI_CLK_PIN     PB5             //      "          13
  #define SPI_SS_PIN      PB2             //      "          10
  #define SPI_RST_PIN     PB2             //      "          10


  #define  spi_read()      ( spi_out(0xff) )
  #define  spi_setmosi()   ( SPI_PORT |= (1 << SPI_MOSI_PIN) )
  #define  spi_clrmosi()   ( SPI_PORT &= ~(1 << SPI_MOSI_PIN) )
  #define  spi_setclk()    ( SPI_PORT |= (1 << SPI_CLK_PIN) )
  #define  spi_clrclk()    ( SPI_PORT &= ~(1 << SPI_CLK_PIN) )
  #define  spi_rdmiso()    ( (SPI_INPUT & (1 << SPI_MISO_PIN)) )
  #define  spi_enable()    ( SPCR |= (1 << SPE) )
  #define  spi_disable()   ( SPCR &= ~(1 << SPE) )

/* ----------------------------------------------------------
   SPI_INIT

     initialisiert die SPI Hardware des AVR Controllers

     speed => Taktgeschwindigkeit SPI-Bus (auch abhaengig
              von clkx2)

                    clkx2 = 0        clkx2 = 1

                0 : F_CPU / 4        F_CPU / 2
                1 : F_CPU / 16       F_CPU / 8
                2 : F_CPU / 64       F_CPU / 32
                3 : F_CPU / 128      F_CPU / 64
   ---------------------------------------------------------- */
void spi_init(uint8_t speed, uint8_t clkx2)
{
  SPI_DDR |= (1 << SPI_MOSI_PIN) | (1 << SPI_CLK_PIN) | (1 << SPI_SS_PIN) | (1 << SPI_RST_PIN);

  SPI_PORT |= (1 << SPI_MISO_PIN);              // Pop-Up Widerstand fuer MISO
  SPI_DDR &= ~(1 << SPI_MISO_PIN);


  // Enable SPI Port: Keine Interrupts, MSB first, Master Modus


  if (clkinv) SPCR= (1 << CPOL);
         else SPCR= 0;

  SPCR |= (1 << SPE) | (1 << MSTR);


  // SPI-Taktrate

  if (speed & 0x01) { SPCR |= (1 << SPR0); }
  if (speed & 0x02) { SPCR |= (1 << SPR1); }

  if (clkx2) { SPSR = (1 << SPI2X); }
        else { SPSR = 0; }
}

/* ----------------------------------------------------------
   SPI_OUT

      Byte ueber SPI senden / empfangen
      data ==> zu sendendes Datum
   ---------------------------------------------------------- */

unsigned char spi_out(unsigned char data)
{
  SPDR=data;                                // Datum senden
  while( ( SPSR & 0x80 ) != 0x80 );         // ... und warten bis es gesendet ist
  return SPDR;                              // empfangenes Byte ist Rueckgabewert
}


void spi_setrst(void)
{
  if (rinv)
    SPI_PORT &= ~(1 << SPI_RST_PIN);
  else
    SPI_PORT |= (1 << SPI_RST_PIN);
}

void spi_clrrst(void)
{
  if (!rinv)
    SPI_PORT &= ~(1 << SPI_RST_PIN);
  else
    SPI_PORT |= (1 << SPI_RST_PIN);
}


/* --------------------------------------------------
    serielle Schnittstelle initialisieren
   -------------------------------------------------- */


/* --------------------------------------------------
    Initialisierung der seriellen Schnittstelle:

    Uebergabe: baud = Baudrate
    Protokoll: 8 Daten-, 1 Stopbit
   -------------------------------------------------- */

void uart_init(uint32_t baud)
{
  uint16_t ubrr;

  if (baud> 57600)
  {
    baud= baud>>1;
    ubrr= (F_CPU/16/baud);
    ubrr--;
    UCSR0A |= 1<<U2X0;                                  // Baudrate verdoppeln
  }
  else
  {
    ubrr= (F_CPU/16/baud-1);
  }
  UBRR0H = (uint8_t)(ubrr>>8);                    // Baudrate setzen
  UBRR0L = (uint8_t)ubrr;

#if defined (__AVR_ATmega328P__) || defined (__AVR_ATmega168__) || defined (__AVR_ATtiny2313__)

  UCSR0B = (1<<RXEN0)|(1<<TXEN0);                       // Transmitter und Receiver enable
  UCSR0C = (3<<UCSZ00);                                 // 8 Datenbit, 1 Stopbit

#else

  // ATmega8
  UCSR0B |= (1<<RXEN0) | (1<<TXEN0);
  UCSRC  = (1<<URSEL) | (1<<UCSZ01) | (1<<UCSZ00);

#endif
}

/* --------------------------------------------------
    Zeichen ueber die serielle Schnittstelle senden
   -------------------------------------------------- */

void uart_putchar(uint8_t ch)
{
  while (!( UCSR0A & (1<<UDRE0)));                      // warten bis Transmitterpuffer leer ist
  UDR0 = ch;                                            // Zeichen senden
}

/* --------------------------------------------------
    testen, ob ein Zeichen auf der Schnittstelle
    ansteht
   -------------------------------------------------- */

uint8_t uart_ischar( void )
{
  return (UCSR0A & (1<<RXC0));
}

/* --------------------------------------------------
    Zeichen von serieller Schnittstelle lesen
   -------------------------------------------------- */

uint8_t uart_getchar( void )
{
  while(!(UCSR0A & (1<<RXC0)));                         // warten bis Zeichen eintrifft

#ifdef send_echo

  char ch;
  ch= UDR0;
  uart_putchar(ch);
  return ch;

#else

  return UDR0;

#endif
}


/* --------------------------------------------------
     RS232_PUTRAMSTRING
     gibt einen String aus dem RAM ueber die RS-232
     aus.
   -------------------------------------------------- */

void uart_putramstring(uint8_t *p)
{
  do
  {
    uart_putchar( *p );
  } while( *p++);
}

/* --------------------------------------------------
     RS232_PUTROMSTRING
     gibt einen String aus dem ROM ueber die RS-232
     aus.

     Bsp.:

          rs232_putromstring(PSTR("Hallo Welt")
   -------------------------------------------------- */

void uart_putromstring(const uint8_t *dataPtr)
{
  uint8_t c;

  for (c=pgm_read_byte(dataPtr); c; ++dataPtr, c=pgm_read_byte(dataPtr))
  {
    uart_putchar(c);
  }
}

/* --------------------------------------------------
     PRINTHEX
     gibt einen unsigned char als Hex-Wert auf der
     seriellen Schnittstelle aus
   -------------------------------------------------- */
void printhex(uint8_t value)
{
  uint8_t b;

  b= (value >> 4);
  if (b< 10) b += '0'; else b += 'A'-10;
  uart_putchar(b);
  b= value & 0x0f;
  if (b< 10) b += '0'; else b += 'A'-10;
  uart_putchar(b);
}

/* --------------------------------------------------
     GETHEX
     liest einen Hex-Wert von der seriellen Schnitt-
     stelle ein

     echo => 1 : gelesener Wert wird zurueckgesendet
             0 : nicht zuruecksenden
   -------------------------------------------------- */
uint8_t gethex(char echo)
{
  uint8_t ch;
  uint8_t value;

  ch= uart_getchar();
  if (echo) uart_putchar(ch);
  if (ch > 'F')
  {
    ch= (ch-'a')+10;
  }
  else
  {
    if (ch> '9') ch= (ch-'A')+10; else ch -= '0';
  }
  value= (ch<< 4);
  ch= uart_getchar();
  if (echo) uart_putchar(ch);

  if (ch > 'F')
  {
    ch= (ch-'a')+10;
  }
  else
  {
    if (ch> '9') ch= (ch-'A')+10; else ch -= '0';
  }
  value |= ch;
  return value;
}

void printhex16(uint16_t value)
{
  printhex(value >> 8);
  printhex(value & 0x00ff);
}

uint16_t gethex16(char echo)
{
  uint16_t i, i2;

  i= gethex(0);
  i= i & 0x00ff;
  i2= gethex(0);
  i2= i2 & 0x00ff;
  i2= i2 | (i << 8);
  if (echo) printhex16(i2);
  return i2;
}


/* --------------------------------------------------
     READSTR
     liest einen String von der seriellen Schnitt-
     stelle ein
   -------------------------------------------------- */
void readstr(char *string, char anz)
{
  char cnt, ch;

  cnt= 0;

  *string= 0;
  do
  {
    ch= uart_getchar();
    if (ch!= 0x0d)
    {
      if (ch!= 0x08)
      {
        if (cnt< anz)
        {
          uart_putchar(ch);
          *string= ch;
          string++;
          *string= 0;
          cnt++;
        }
      }
      else
      {
        if (cnt)
        {
          uart_putchar(0x08);
          string--;
          *string= 0;
          cnt--;
        }
      }
    }
  } while (ch != 0x0d);
}

/* --------------------------------------------------
     CRLF
     erzeugt einen Carriagereturn und linefeed auf
     der seriellen Schnittstelle
   -------------------------------------------------- */

void crlf(void)
{
  uart_putchar(0x0d);
  uart_putchar(0x0a);
}




char checkcmd(char *string)
{
  char  i,cp, ch;
  char  match;
  char  *s;

  i= 0;
  do
  {
    s= string;
    match= 1;
    cp= 0;
    do
    {
      ch= pgm_read_byte(&cmds[i][cp]);
      if (ch!= *s) match= 0;
//      printf("%d: %.2x = %c  %.2x = %c\n\r",cp,ch,ch,*s,*s);
      cp++; s++;
    }while(ch && match);
    if (match) return i+1;
    i++;
  } while (i < cmdanz);
  return 0;
}


void showactrstmode(void)
{
  crlf();
  uart_prints("actual mode is: ");
  if (rinv) uart_prints("invert "); else uart_prints("not invert ");
}

void showactclkmode(void)
{
  crlf();
  uart_prints("actual mode is: ");
  if (clkinv) uart_prints("invert "); else uart_prints("not invert ");
}

void get3cmdbytes(uint8_t *v1, uint8_t *v2, uint8_t *v3)
{
  crlf();
  uart_prints("Enter 3 instruction bytes: ");
  *v1 = gethex(1); uart_putchar(' ');
  *v2 = gethex(1); uart_putchar(' ');
  *v3 = gethex(1); uart_putchar(' ');
  crlf();
}

void get4cmdbytes(uint8_t *v1, uint8_t *v2, uint8_t *v3, uint8_t *v4)
{
  crlf();
  uart_prints("Enter 4 instruction bytes: ");
  *v1 = gethex(1); uart_putchar(' ');
  *v2 = gethex(1); uart_putchar(' ');
  *v3 = gethex(1); uart_putchar(' ');
  *v4 = gethex(1); uart_putchar(' ');
  crlf();
}

void send3cmdbytes(uint8_t v1, uint8_t v2, uint8_t v3)
{
  spi_out(v1);
  spi_out(v2);
  spi_out(v3);
}

uint8_t send4cmdbytes(uint8_t v1, uint8_t v2, uint8_t v3, uint8_t v4)
{
  uint8_t value;

  spi_out(v1);
  spi_out(v2);
  spi_out(v3);
  value= spi_out(v4);
  return value;
}

void spi_setfreq(uint8_t value)
{
  spi_disable();
  switch (value)
  {
    case 1 :   spi_init(3,0); break;          //    F_CPU / 128
    case 2 :   spi_init(2,0); break;          //    F_CPU / 64
    case 3 :   spi_init(2,1); break;          //    F_CPU / 32
    case 4 :   spi_init(1,0); break;          //    F_CPU / 16
    case 5 :   spi_init(1,1); break;          //    F_CPU / 8
    case 6 :   spi_init(0,0); break;          //    F_CPU / 4
    case 7 :   spi_init(0,1); break;          //    F_CPU / 2
    default : break;
  }
  aktfreq= value;
}

void dump(void)
{
  uint16_t startadr, stopadr;
  uint16_t cnt, pcnt;
  uint8_t buffer[8];
  uint8_t bcx, ch, i;

  uart_prints("\n\rStartadress: ");
  ch= gethex(1);
  startadr= (ch << 8);
  ch= gethex(1);
  startadr |= ch;

  uart_prints("\n\rStopadress : ");
  ch= gethex(1);
  stopadr= (ch << 8);
  ch= gethex(1);
  stopadr |= ch;
  crlf();
  crlf();
  cnt= startadr;
  pcnt= cnt;
  bcx= 0;
  for (i= 0; i<8; i++) buffer[i]= 0;
  do
  {
    send3cmdbytes(0x20, (cnt >> 8), (cnt & 0x00ff));

    buffer[bcx]= spi_read();
    cnt++; bcx++;
    if (bcx> 7)
    {
      printhex(pcnt >> 8); printhex(pcnt & 0x00ff);
      pcnt= cnt;
      uart_prints(": ");
      for (i= 0; i< 8; i++)
      {
        printhex(buffer[i]);
        uart_putchar(' ');
      }
      uart_prints("   ");
      for (i= 0; i< 8; i++)
      {
        if ((buffer[i]> 31) && (buffer[i]< 128))
          uart_putchar(buffer[i]);
        else
          uart_putchar('.');
      }
      crlf();
      bcx= 0;
    }
  } while(cnt < stopadr+1);
  if (bcx)
  {
    printhex(pcnt >> 8); printhex(pcnt & 0x00ff);
    uart_prints(": ");
    for (i= 0; i<8; i++)
    {
      if (i< bcx)
      {
        printhex(buffer[i]);
        uart_putchar(' ');
      }
      else
      {
        uart_prints("   ");
      }
    }
    uart_prints("   ");
    for (i= 0; i<8; i++)
    {
      if (i< bcx)
      {
        if ((buffer[i]> 31) && (buffer[i]< 128))
          uart_putchar(buffer[i]);
      }
    }
  }
  crlf();
}

void cmdinterpret(void)
{
  char ch;
  uint8_t b;
  uint8_t v1, v2, v3, v4;
  char inbuffer[13];

  uart_prints("\n\rSPI-evaluation\n\rexplore a SPI-Device\n\n\r");

  crlf();

  do
  {
    uart_prints("SPI:> ");
    readstr(&inbuffer[0],8);
    ch= checkcmd(&inbuffer[0]);
    switch(ch)
    {
      case 1 :                                // Hardware SPI disable
      {
        spi_disable();
        uart_prints("\n\rSPI disabled\n\r");
        break;
      }
      case 2 :                                // Hardware SPI enable
      {
        uart_prints("\n\rSPI enabled\n\r");
        spi_setfreq(aktfreq);
        break;
      }
      case 3 :                                // Mosi - Leitung
      {
        crlf();
        uart_prints("[0/1]: ");
        ch= uart_getchar();

        uart_putchar(ch);
        if (ch== '1') { spi_setmosi(); } else { spi_clrmosi(); }
        crlf();
        break;
      }
      case 4 :                                // Miso - Leitung lesen
      {
        crlf();
        if (spi_rdmiso()) {uart_prints("Miso= 1\n\r"); } else {uart_prints("Miso= 0\n\r");}
        break;
      }
      case 5 :                                // Reset - Leitung
      {
        crlf();
        uart_prints("[0/1]: ");
        ch= uart_getchar();
        uart_putchar(ch);
        if (ch== '1') { spi_setrst(); } else { spi_clrrst(); }
        crlf();
        break;
      }
      case 6 :                                // SCK (Clock) - Leitung
      {
        crlf();
        uart_prints("[0/1]: ");
        ch= uart_getchar();
        uart_putchar(ch);
        if (ch== '1') { spi_setclk(); } else { spi_clrclk(); }
        crlf();
        break;
      }
      case 7 :                               // Reset-Modus (invertiert oder nicht invertiert)
      {
        showactrstmode();
        uart_prints("[0/1]: ");
        ch= uart_getchar();
        uart_putchar(ch);
        if (ch== '1') {rinv= 1;} else {rinv= 0;}
        showactrstmode();
        crlf();
        break;
      }
      case 8 :                               // send Data-Byte
      {
        crlf();
        uart_prints("byte to send: ");
        b= gethex(1);
        spi_out(b);
        crlf();
        break;
      }
      case 9 :                               // read Data-Byte
      {
        crlf();
        uart_prints("readed byte: ");
        b= spi_read();
        printhex(b);
        crlf();
        break;
      }
      case 10 :                             // send 3-Byte instruction
      {
        get3cmdbytes(&v1, &v2, &v3);
        send3cmdbytes(v1, v2, v3);
        break;
      }
      case 11:                              // SPI clockfrequency
      {
        crlf();
        uart_prints("[1 .. 7]: ");
        ch= uart_getchar();
        uart_putchar(ch);
        spi_setfreq(ch-'0');
        crlf();
        break;
      }
      case 12:
      {
        dump();
        break;
      }
      case 13 :                               // Clock-Modus (invertiert oder nicht invertiert)
      {
        showactclkmode();
        uart_prints("[0/1]: ");
        ch= uart_getchar();
        uart_putchar(ch);
        if (ch== '1') {clkinv= 1;} else {clkinv= 0;}
        showactclkmode();
        spi_setfreq(aktfreq);
        crlf();
        break;
      }
      case 14 :                             // send 4-Byte instruction
      {
        get4cmdbytes(&v1, &v2, &v3, &v4);
        send4cmdbytes(v1, v2, v3, v4);
        break;
      }
      case 15 : uart_prints("\n\r"); break; // exit cmd-mode


/*      case 'x' :
        {
        } */
      default :
        {
          uart_putchar(ch);
          uart_prints("\n\runkown command\n\r");
          break;
        }
    }
  }while(ch != 15);
}

/*
   Reset Target und initialisiert program enable
   gibt angeschlossenen Controller zurueck
   ID-Nr:   73 = AT89S8253
            43 = AT89S4051
            23 = AT89S2051
            52 = AT89S52
            51 = AT89S51
            69 = ID defekt, MCU ist entweder 51 oder 52
*/

uint8_t progmode_init(void)
{
  uint8_t id;

  id= 0xff;
  // Chip reset
  clkinv= 0;
  spi_setfreq(aktfreq);
  spi_setrst();
  _delay_ms(15);
  spi_clrrst();
  _delay_ms(15);
  spi_setrst();
  _delay_ms(15);

  id= send4cmdbytes(0xac, 0x53, 0x00, 0x00);       // program enable command
  if (id== 0x69)
  {                                                // es ist ein S51 oder S52
    id= send4cmdbytes(0x28, 0x01, 0x00, 0x00);     // read ID command
    if (!(id == 0x51) && !(id |= 0x52))
    {
      id= 0x69;
      return id;
    }
    return id;
    uart_putchar('p');                             // sollte nie gesendet werden ...
  }
  id= 0xff;
  clkinv= 1;                                       // AT89S8253 und AT89Sx051 haben
  spi_setfreq(aktfreq);                            // invertierte Clockuebernahme
  spi_setrst();
  _delay_ms(15);
  spi_clrrst();
  _delay_ms(15);
  spi_setrst();
  _delay_ms(15);

  id= send4cmdbytes(0xac, 0x53, 0x00, 0x00);       // program enable command
  _delay_ms(4);
  id= send4cmdbytes(0x28, 0x00, 0x31, 0x00);       // read ID command S8253
  if (id== 0x73) return id;                        // es ist ein S8253
  id= send4cmdbytes(0x28, 0x00, 0x01, 0x00);       // read ID command Sx051
  if ((id== 0x43) || (id== 0x23)) return id;         // es ist ein Sx051

  return 0xff;
}

void mcu_readflash(void)
{
  uint8_t b, id;
  uint16_t cnt, proglen;

  id= progmode_init();

  switch (id)
  {
    case 0x69  : { proglen= 0x2000; break; }       // 89S51 oder S52, vorsichtshalber 8 kByte Flash
    case 0x51  : { proglen= 0x1000; break; }       // 89S51   =>  4 kByte Flash
    case 0x52  : { proglen= 0x2000; break; }       // 89S51   =>  8 kByte Flash
    case 0x73  : { proglen= 0x3000; break; }       // 89S8253 => 12 KByte Flash
    case 0x23  : { proglen= 0x0800; break; }       // 89S2051 =>  2 KByte Flash
    case 0x43  : { proglen= 0x1000; break; }       // 89S8253 =>  4 KByte Flash
    default: proglen= 0x0000; break;
  }

  printhex(proglen>> 8);
  printhex(proglen & 0x00ff);

  if (!proglen) return;

  for (cnt= 0; cnt< proglen; cnt++)
  {
    send3cmdbytes(0x20, (cnt >> 8), (cnt & 0x00ff));
    b= spi_read();
    printhex(b);
//    _delay_us(10);
  }

}


void mcu_reset(void)
{
  spi_clrrst();
  _delay_ms(15);
  spi_setrst();
  _delay_ms(15);
  spi_clrrst();
  _delay_ms(15);
  uart_putchar('r');
}

void mcu_erase(void)
{
  uint8_t b;

  uart_putchar('e');
  send4cmdbytes(0xac, 0x80, 0x00, 0x00);           // erase chip
  _delay_ms(e_delay);
}

void mcu_flash(void)
{
  uint16_t cx;
  uint8_t b, id;

  id= progmode_init();

  if ((id== 0x51) || (id== 0x52) || (id== 0x69))   // AT89S51 / S52
  {
    for (cx= 0; cx< flashbytes+1; cx++)
    {
      b= gethex(0);
      send4cmdbytes(0x40,cx >> 8, cx & 0x00ff,b);  // Bytewritemodus
      _delay_us(b_delay);
      printhex(b);                                 // Echo Hexzahl (Ascii)
    }
  }

  if (id== 0x73)                                   // AT89S8253
  {
    for (cx= 0; cx< flashbytes+1; cx++)
    {
      b= gethex(0);
      if ((cx % page8253) == 0)                       // Speichergroesse einer Page AT89S8253
      {
        _delay_us(p_delay);                           // Schreibdauer einer Page
        send4cmdbytes(0x50, cx>> 8, cx & 0x00ff,b);   // Pageadresse und Byte 0 der Page
      }
      else
      {
        spi_out(b);
      }
      printhex(b);
    }
    if (cx % page8253)                                // wenn eine Page nicht komplett ist
    {
      while(cx % page8253)                            // dann Speicherzellen mit 0xFF auffuellen
      {
        spi_out(0xff);
        cx++;
      }
    }
  }

  if ((id== 0x23) || (id== 0x43))                     // AT89S2051 / S4051
  {


//  nachfolgender Code funktioniert warum auch immer nicht: fix it !
//  26.04.2016  17:15 scheint zu funktionieren, abschliesende Wartezeit hat
//  gefehlt

    for (cx= 0; cx< flashbytes+1; cx++)
    {
      b= gethex(0);
      if ((cx % page2051) == 0)                       // Speichergroesse einer Page AT89S2051
      {
        _delay_us(p_delay2);                          // Schreibdauer einer Page
        send4cmdbytes(0x50, cx>> 8, cx & 0x00ff,b);   // Pageadresse und Byte 0 der Page
      }
      else
      {
        spi_out(b);
      }
      printhex(b);
    }
    if (cx % page2051)                                // wenn eine Page nicht komplett ist
    {
      while(cx % page2051)                            // dann Speicherzellen mit 0xFF auffuellen
      {
        spi_out(0xff);
        cx++;
      }
      _delay_us(p_delay2);                          // Schreibdauer einer Page
    }

/*
//  Bytewritemodus... wenn nix gehen sollte

    for (cx= 0; cx< flashbytes+1; cx++)
    {
      b= gethex(0);
      send4cmdbytes(0x40,cx >> 8, cx & 0x00ff,b);  // Bytewritemodus
      _delay_us(b_delay2);
      printhex(b);                                 // Echo Hexzahl (Ascii)
    }
*/

  }
}

void uartbridge_init(void)
{
  serportddr |= (1 << sertxd) | (1 << seroutrxd);
  serportddr &= ~( (1 << serrxd | (1 << serintxd) ));
  serport |= (1 << sertxd) | (1 << serrxd) | (1 << serintxd);   // txd= 1 , Pull-Up Widerstand rxd an

  button_init();
}

/* --------------------------------------------------

                       M A I N

   -------------------------------------------------- */

int main(void)
{
  char       ch;
  uint8_t    b;
  uint8_t    v1, v2, v3, v4;
  char       inbuffer[13];
  uint16_t   i;

  uart_init(115200);
  uartbridge_init();
  led_ge_init();
  led_gn_init();
  led_rt_init();

  led_ge_set();
  led_rt_clr();
  led_gn_clr();                                                   // Programmer ist betriebsbereit

//  spi_init(2,1);
//  spi_disable();

  aktfreq= 2; clkinv= 0;
  spi_setfreq(aktfreq);

  while(1)
  {
    if (!(is_button()) )                                          // UART-Button nicht gedrueckt
    {
      if (uart_ischar() )
      {
        ch= uart_getchar();
        switch (ch)
        {
          case '1' :  mcutyp= progmode_init(); mcu_erase(); break;
          case '3' :  flashbytes= gethex16(1); break;               // Anzahl zu flashender Bytes
          case '4' :  mcu_reset(); break;
          case '5' :  cmdinterpret(); break;
          case '6' :  {
                        led_gn_set();
                        mcutyp= progmode_init();
                        mcu_flash();
                        led_gn_clr();
                        break;
                      }
          case '7' :  mcu_readflash(); break;

          case '8' :
          {
            mcutyp= progmode_init();
            printhex(mcutyp);
            break;
          }

          case '9' :
          {
            printhex(0x18);
            break;
          }

          default : break;
        }
      }
    }
    else                                                          // Bruecke MCS51 zu RS232 via adapter
    {
      _delay_ms(50);                                              // entprellen
      while(is_button() );                                        // bis Button wieder losgelassen wird
      _delay_ms(50);
      UCSR0B &= ~((1<<RXEN0) | (1<<TXEN0));                       // Transmitter und Receiver off (Hardware UART aus)

      led_rt_set();                                               // Bridge ist aktiv
      led_ge_clr();
      while(!(is_button()) )
      {
        if (isserintxd()) { settxd(); } else { clrtxd(); }
        if (isrxd()) { setoutrxd(); } else { clroutrxd(); }
      }

      _delay_ms(50);                                              // entprellen
      while(is_button() );                                      // bis Button wieder losgelassen wird
      _delay_ms(50);
      UCSR0B = (1<<RXEN0)|(1<<TXEN0);                             // Transmitter und Receiver enable (Hardware UART wieder an)
      led_rt_clr();                                               // Bridge ist aktiv
      led_ge_set();
    }
  }
}
